@@ -4,7 +4,7 @@ use strict;
 BEGIN {
 	require 5.004;
 	require Exporter;
-	$YAML::Tiny::VERSION   = '1.21';
+	$YAML::Tiny::VERSION   = '1.23';
 	$YAML::Tiny::errstr    = '';
 	@YAML::Tiny::ISA       = qw{ Exporter  };
 	@YAML::Tiny::EXPORT_OK = qw{
@@ -14,36 +14,22 @@ BEGIN {
 		};
 }
 
-# Create the main error hash
-my %ERROR = (
-	YAML_PARSE_ERR_NO_FINAL_NEWLINE => "Stream does not end with newline character",
-	
-);
-
-my %NO = (
-	'%' => 'YAML::Tiny does not support directives',
-	'&' => 'YAML::Tiny does not support anchors',
-	'*' => 'YAML::Tiny does not support aliases',
-	'?' => 'YAML::Tiny does not support explicit mapping keys',
-	':' => 'YAML::Tiny does not support explicit mapping values',
-	'!' => 'YAML::Tiny does not support explicit tags',
-);
-
 my $ESCAPE_CHAR = '[\\x00-\\x08\\x0b-\\x0d\\x0e-\\x1f\"\n]';
 
 # Escapes for unprintable characters
-my @UNPRINTABLE = qw(z    x01  x02  x03  x04  x05  x06  a
-                     x08  t    n    v    f    r    x0e  x0f
-                     x10  x11  x12  x13  x14  x15  x16  x17
-                     x18  x19  x1a  e    x1c  x1d  x1e  x1f
-                    );
+my @UNPRINTABLE = qw(
+	z    x01  x02  x03  x04  x05  x06  a
+	x08  t    n    v    f    r    x0e  x0f
+	x10  x11  x12  x13  x14  x15  x16  x17
+	x18  x19  x1a  e    x1c  x1d  x1e  x1f
+);
 
 # Printable characters for escapes
 my %UNESCAPES = (
 	z => "\x00", a => "\x07", t    => "\x09",
 	n => "\x0a", v => "\x0b", f    => "\x0c",
 	r => "\x0d", e => "\x1b", '\\' => '\\',
-	);
+);
 
 # Create an empty YAML::Tiny object
 sub new {
@@ -63,9 +49,10 @@ sub read {
 
 	# Slurp in the file
 	local $/ = undef;
-	open CFG, $file or return $class->_error( "Failed to open file '$file': $!" );
+	local *CFG;
+	open( CFG, $file ) or return $class->_error( "Failed to open file '$file': $!" );
 	my $contents = <CFG>;
-	close CFG;
+	close( CFG ) or return $class->_error( "Failed to close file '$file': $!" );
 
 	$class->read_string( $contents );
 }
@@ -79,7 +66,7 @@ sub read_string {
 	return undef unless defined $_[0];
 	return $self unless length $_[0];
 	unless ( $_[0] =~ /[\012\015]+$/ ) {
-		return $class->_error('YAML_PARSE_ERR_NO_FINAL_NEWLINE');
+		return $class->_error("Stream does not end with newline character");
 	}
 
 	# Split the file into lines
@@ -115,19 +102,13 @@ sub read_string {
 			$self->_read_hash( $document, [ length($1) ], \@lines );
 
 		} else {
-			die "CODE INCOMPLETE";
+			die "YAML::Tiny does not support the line '$lines[0]'";
 		}
 	}
 
 	$self;
 }
 
-sub _check_support {
-	# Check if we support the next char
-	my $err = $NO{substr($_[1], 0, 1)};
-	Carp::croak($err) if $err;
-}
-
 # Deparse a scalar string to the actual scalar
 sub _read_scalar {
 	my ($self, $string, $indent, $lines) = @_;
@@ -139,13 +120,13 @@ sub _read_scalar {
 	return undef if $string eq '~';
 
 	# Quotes
-	if ( $string =~ /^'(.*?)'$/ ) {
+	if ( $string =~ /^\'(.*?)\'$/ ) {
 		return '' unless defined $1;
 		my $rv = $1;
-		$rv =~ s/''/'/g;
+		$rv =~ s/\'\'/\'/g;
 		return $rv;
 	}
-	if ( $string =~ /^"((?:\\.|[^"])*)"$/ ) {
+	if ( $string =~ /^\"((?:\\.|[^\"])*)\"$/ ) {
 		my $str = $1;
 		$str =~ s/\\"/"/g;
 		$str =~ s/\\([never\\fartz]|x([0-9a-fA-F]{2}))/(length($1)>1)?pack("H2",$2):$UNESCAPES{$1}/gex;
@@ -208,7 +189,7 @@ sub _read_array {
 			die "Hash line over-indented";
 		}
 
-		if ( $lines->[0] =~ /^(\s*\-\s+)[^'"]\S*\s*:(?:\s+|$)/ ) {
+		if ( $lines->[0] =~ /^(\s*\-\s+)[^\'\"]\S*\s*:(?:\s+|$)/ ) {
 			# Inline nested hash
 			my $indent2 = length("$1");
 			$lines->[0] =~ s/-/ /;
@@ -242,11 +223,11 @@ sub _read_array {
 				$self->_read_hash( $array->[-1], [ @$indent, length("$1") ], $lines );
 
 			} else {
-				die "CODE INCOMPLETE";
+				die "YAML::Tiny does not support the line '$lines->[0]'";
 			}
 
 		} else {
-			die "CODE INCOMPLETE";
+			die "YAML::Tiny does not support the line '$lines->[0]'";
 		}
 	}
 
@@ -270,7 +251,7 @@ sub _read_hash {
 		}
 
 		# Get the key
-		unless ( $lines->[0] =~ s/^\s*([^'"][^\n]*?)\s*:(\s+|$)// ) {
+		unless ( $lines->[0] =~ s/^\s*([^\'\"][^\n]*?)\s*:(\s+|$)// ) {
 			die "Bad hash line";
 		}
 		my $key = $1;
@@ -347,7 +328,7 @@ sub write_string {
 				$lines[-1] .= ' []';
 				next;
 			}
-			push @lines, $self->_write_array( $indent, $cursor );
+			push @lines, $self->_write_array( $cursor, $indent, {} );
 
 		# A hash at the root
 		} elsif ( ref $cursor eq 'HASH' ) {
@@ -355,7 +336,7 @@ sub write_string {
 				$lines[-1] .= ' {}';
 				next;
 			}
-			push @lines, $self->_write_hash( $indent, $cursor );
+			push @lines, $self->_write_hash( $cursor, $indent, {} );
 
 		} else {
 			Carp::croak("Cannot serialize " . ref($cursor));
@@ -376,41 +357,45 @@ sub _write_scalar {
 		return qq{"$str"};
 	}
 	if ( length($str) == 0 or $str =~ /\s/ ) {
-		$str =~ s/'/''/;
+		$str =~ s/\'/\'\'/;
 		return "'$str'";
 	}
 	return $str;
 }
 
 sub _write_array {
-	my ($self, $indent, $array) = @_;
+	my ($self, $array, $indent, $seen) = @_;
+	if ( $seen->{refaddr($array)}++ ) {
+		die "YAML::Tiny does not support circular references";
+	}
 	my @lines  = ();
 	foreach my $el ( @$array ) {
 		my $line = ('  ' x $indent) . '-';
-		if ( ! ref $el ) {
+		my $type = ref $el;
+		if ( ! $type ) {
 			$line .= ' ' . $self->_write_scalar( $el );
 			push @lines, $line;
 
-		} elsif ( ref $el eq 'ARRAY' ) {
+		} elsif ( $type eq 'ARRAY' ) {
 			if ( @$el ) {
 				push @lines, $line;
-				push @lines, $self->_write_array( $indent + 1, $el );
+				push @lines, $self->_write_array( $el, $indent + 1, $seen );
 			} else {
 				$line .= ' []';
 				push @lines, $line;
 			}
 
-		} elsif ( ref $el eq 'HASH' ) {
+		} elsif ( $type eq 'HASH' ) {
 			if ( keys %$el ) {
 				push @lines, $line;
-				push @lines, $self->_write_hash( $indent + 1, $el );
+				push @lines, $self->_write_hash( $el, $indent + 1, $seen );
 			} else {
 				$line .= ' {}';
 				push @lines, $line;
 			}
 
 		} else {
-			die "CODE INCOMPLETE";
+			die "YAML::Tiny does not support $type references";
 		}
 	}
 
@@ -418,35 +403,39 @@ sub _write_array {
 }
 
 sub _write_hash {
-	my ($self, $indent, $hash) = @_;
+	my ($self, $hash, $indent, $seen) = @_;
+	if ( $seen->{refaddr($hash)}++ ) {
+		die "YAML::Tiny does not support circular references";
+	}
 	my @lines  = ();
 	foreach my $name ( sort keys %$hash ) {
 		my $el   = $hash->{$name};
 		my $line = ('  ' x $indent) . "$name:";
-		if ( ! ref $el ) {
+		my $type = ref $el;
+		if ( ! $type ) {
 			$line .= ' ' . $self->_write_scalar( $el );
 			push @lines, $line;
 
-		} elsif ( ref $el eq 'ARRAY' ) {
+		} elsif ( $type eq 'ARRAY' ) {
 			if ( @$el ) {
 				push @lines, $line;
-				push @lines, $self->_write_array( $indent + 1, $el );
+				push @lines, $self->_write_array( $el, $indent + 1, $seen );
 			} else {
 				$line .= ' []';
 				push @lines, $line;
 			}
 
-		} elsif ( ref $el eq 'HASH' ) {
+		} elsif ( $type eq 'HASH' ) {
 			if ( keys %$el ) {
 				push @lines, $line;
-				push @lines, $self->_write_hash( $indent + 1, $el );
+				push @lines, $self->_write_hash( $el, $indent + 1, $seen );
 			} else {
 				$line .= ' {}';
 				push @lines, $line;
 			}
 
 		} else {
-			Carp::croak("Cannot serialize " . ref($el));
+			die "YAML::Tiny does not support $type references";
 		}
 	}
 
@@ -455,7 +444,7 @@ sub _write_hash {
 
 # Set error
 sub _error {
-	$YAML::Tiny::errstr = $ERROR{$_[1]} ? "$ERROR{$_[1]} ($_[1])" : $_[1];
+	$YAML::Tiny::errstr = $_[1];
 	undef;
 }
 
@@ -497,6 +486,38 @@ sub LoadFile {
 	return @$self;
 }
 
+
+
+
+
+#####################################################################
+# Use Scalar::Util if possible, otherwise emulate it
+
+BEGIN {
+	eval {
+		require Scalar::Util;
+	};
+	if ( $@ ) {
+		# Failed to load Scalar::Util
+		eval <<'END_PERL';
+sub refaddr {
+	my $pkg = ref($_[0]) or return undef;
+	if (!!UNIVERSAL::can($_[0], 'can')) {
+		bless $_[0], 'Scalar::Util::Fake';
+	} else {
+		$pkg = undef;
+	}
+	"$_[0]" =~ /0x(\w+)/;
+	my $i = do { local $^W; hex $1 };
+	bless $_[0], $pkg if defined $pkg;
+	$i;
+}
+END_PERL
+	} else {
+		Scalar::Util->import('refaddr');
+	}
+}
+
 1;
 
 __END__
